/*
 * latency -- measure the scheduling latency of a particular process from
 *	the extra information provided in /proc/<pid>stat by version 4 of
 *	the schedstat patch. PLEASE NOTE: This program does NOT check to
 *	make sure that extra information is there; it assumes the last
 *	three fields in that line are the ones it's interested in.  Using
 *	it on a kernel that does not have the schedstat patch compiled in
 *	will cause it to happily produce bizarre results.
 *
 *	Note too that this is known to work only with versions 4 and 5
 *	of the schedstat patch, for similar reasons.
 *
 *	This currently monitors only one pid at a time but could easily
 *	be modified to do more.
 *
 *      Modifications licensed under GPLv2
 *      (C) 2009 The Linux Foundation, Christoph Lameter <cl@linux-foundation.org>
 */
#define _ISOC99_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

extern char *index(), *rindex();
char procbuf[512];
char statname[64];
char *Progname;
FILE *fp;

void usage()
{
    fprintf(stderr,"Usage: %s [-s sleeptime ] <pid>\n", Progname);
    exit(-1);
}

/*
 * get_stats() -- we presume that we are interested in the first three
 *	fields of the line we are handed, and further, that they contain
 *	only numbers and single spaces.
 */
void get_stats(char *buf, unsigned long long *run_ns, unsigned long long *wait_ns,
    unsigned long long *nran)
{
    char *ptr;

    /* sanity */
    if (!buf || !run_ns || !wait_ns || !nran)
	return;

    /* leading blanks */
    ptr = buf;
    while (*ptr && isblank(*ptr))
	ptr++;

    /* first number -- run_ticks */
    *run_ns = atoll(ptr);
    while (*ptr && isdigit(*ptr))
	ptr++;
    while (*ptr && isblank(*ptr))
	ptr++;

    /* second number -- wait_ticks */
    *wait_ns = atoll(ptr);
    while (*ptr && isdigit(*ptr))
	ptr++;
    while (*ptr && isblank(*ptr))
	ptr++;

    /* last number -- nran */
    *nran = atoll(ptr);
}

/*
 * get_id() -- extract the id field from that /proc/<pid>/stat file
 */
void get_id(char *buf, char *id)

{
    char *ptr;

    /* sanity */
    if (!buf || !id)
	return;

    ptr = index(buf, ')') + 1;
    *ptr = 0;
    strcpy(id, buf);
    *ptr = ' ';

    return;
}

int main(int argc, char *argv[])
{
    int c;
    unsigned int sleeptime = 5, pid = 0, verbose = 0;
    char id[32];
    unsigned long long run_ns, wait_ns, nran;
    unsigned long long orun_ns=0, owait_ns=0, oran=0;

    Progname = argv[0];
    id[0] = 0;
    while ((c = getopt(argc,argv,"s:v")) != -1) {
	switch (c) {
	    case 's':
		sleeptime = atoi(optarg);
		break;
	    case 'v':
		verbose++;
		break;
	    default:
		usage();
	}
    }

    if (optind < argc) {
	pid = atoi(argv[optind]);
    }

    if (!pid)
	usage();

    sprintf(statname,"/proc/%d/stat", pid);
    if ((fp = fopen(statname, "r"))) {
	if (fgets(procbuf, sizeof(procbuf), fp))
	    get_id(procbuf,id);
	fclose(fp);
    }

    /*
     * now just spin collecting the stats
     */
    sprintf(statname,"/proc/%d/schedstat", pid);

    printf("latencystat v0.1 interval=%d seconds, pid=%s\n", sleeptime, id);

    fp = fopen(statname, "r");
    if (fp) {
	    if (fgets(procbuf, sizeof(procbuf), fp)) {
		get_stats(procbuf, &orun_ns, &owait_ns, &oran);
	    	sleep(sleeptime);
	   } else
		abort();
    }

    fp = fopen(statname, "r");
    while (fp) {
	    unsigned long tran;

	    if (!fgets(procbuf, sizeof(procbuf), fp))
		    break;

	    get_stats(procbuf, &run_ns, &wait_ns, &nran);
	    tran = nran - oran;

	    if (verbose)
		printf("%s %lld(%lld) %lld(%lld) %lld(%lld) %4.2f %4.2f\n",
		    id, run_ns, run_ns - orun_ns,
		    wait_ns, wait_ns - owait_ns,
		    nran, oran,
		    tran ?
			(double)(run_ns-orun_ns)/tran : 0,
		    tran ?
			(double)(wait_ns-owait_ns)/tran : 0);
	    else if (tran)
		printf("%s %ld times scheduled avgrun=%4.2fus avgwait=%4.2fus\n",
		    id, tran,
			(double)(run_ns-orun_ns)/ (tran * 1000.0),
			(double)(wait_ns-owait_ns) / (tran * 1000.0));
	    else
		printf("%s was not executing in the observation interval\n", id);

	    fclose(fp);
	    oran = nran;
	    orun_ns = run_ns;
	    owait_ns = wait_ns;
	    sleep(sleeptime);
	    fp = fopen(statname,"r");
	    if (!fp)
		    break;
    }
    if (id[0])
	printf("Process %s has exited.\n", id);
    else 
	printf("Process %d does not exist.\n", pid);
    exit(0);
}
